@@ -43,8 +43,6 @@ enum connman_device_type {
 	CONNMAN_DEVICE_TYPE_BLUETOOTH = 4,
 	CONNMAN_DEVICE_TYPE_CELLULAR  = 5,
 	CONNMAN_DEVICE_TYPE_GPS       = 6,
-	CONNMAN_DEVICE_TYPE_MBM       = 14,
-	CONNMAN_DEVICE_TYPE_HSO       = 15,
 	CONNMAN_DEVICE_TYPE_VENDOR    = 10000,
 };
 
@@ -81,7 +79,6 @@ int connman_device_set_powered(struct connman_device *device,
 						connman_bool_t powered);
 int connman_device_set_scanning(struct connman_device *device,
 						connman_bool_t scanning);
-void connman_device_cleanup_scanning(struct connman_device *device);
 
 int connman_device_set_disconnected(struct connman_device *device,
 						connman_bool_t disconnected);
