@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -33,6 +33,7 @@
 
 #define CONNMAN_API_SUBJECT_TO_CHANGE
 #include <connman/plugin.h>
+#include <connman/technology.h>
 #include <connman/device.h>
 #include <connman/inet.h>
 #include <connman/dbus.h>
@@ -62,6 +63,7 @@
 static DBusConnection *connection;
 
 static GHashTable *bluetooth_devices = NULL;
+static GHashTable *bluetooth_networks = NULL;
 
 static int pan_probe(struct connman_network *network)
 {
@@ -86,11 +88,15 @@ static void connect_reply(DBusPendingCall *call, void *user_data)
 	DBG("network %p", network);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	dbus_error_init(&error);
 
+	if (dbus_set_error_from_message(&error, reply) == TRUE) {
+		connman_error("%s", error.message);
+		dbus_error_free(&error);
+		goto done;
+	}
+
 	if (dbus_message_get_args(reply, &error,
 					DBUS_TYPE_STRING, &interface,
 						DBUS_TYPE_INVALID) == FALSE) {
@@ -110,10 +116,13 @@ static void connect_reply(DBusPendingCall *call, void *user_data)
 	index = connman_inet_ifindex(interface);
 
 	connman_network_set_index(network, index);
+
 	connman_network_set_connected(network, TRUE);
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static int pan_connect(struct connman_network *network)
@@ -128,9 +137,6 @@ static int pan_connect(struct connman_network *network)
 	if (path == NULL)
 		return -EINVAL;
 
-	if (connman_network_get_index(network) >= 0)
-		return -EISCONN;
-
 	message = dbus_message_new_method_call(BLUEZ_SERVICE, path,
 					BLUEZ_NETWORK_INTERFACE, CONNECT);
 	if (message == NULL)
@@ -170,11 +176,15 @@ static void disconnect_reply(DBusPendingCall *call, void *user_data)
 	DBG("network %p", network);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	dbus_error_init(&error);
 
+	if (dbus_set_error_from_message(&error, reply) == TRUE) {
+		connman_error("%s", error.message);
+		dbus_error_free(&error);
+		goto done;
+	}
+
 	if (dbus_message_get_args(reply, &error, DBUS_TYPE_INVALID) == FALSE) {
 		if (dbus_error_is_set(&error) == TRUE) {
 			connman_error("%s", error.message);
@@ -185,10 +195,13 @@ static void disconnect_reply(DBusPendingCall *call, void *user_data)
 	}
 
 	connman_network_set_connected(network, FALSE);
-	connman_network_set_index(network, -1);
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
+
+	connman_network_unref(network);
 }
 
 static int pan_disconnect(struct connman_network *network)
@@ -202,9 +215,6 @@ static int pan_disconnect(struct connman_network *network)
 	if (path == NULL)
 		return -EINVAL;
 
-	if (connman_network_get_index(network) < 0)
-		return -ENOTCONN;
-
 	message = dbus_message_new_method_call(BLUEZ_SERVICE, path,
 					BLUEZ_NETWORK_INTERFACE, DISCONNECT);
 	if (message == NULL)
@@ -227,6 +237,10 @@ static int pan_disconnect(struct connman_network *network)
 		return -EINVAL;
 	}
 
+	connman_network_ref(network);
+
+	connman_network_set_associating(network, FALSE);
+
 	dbus_pending_call_set_notify(call, disconnect_reply, network, NULL);
 
 	dbus_message_unref(message);
@@ -243,99 +257,43 @@ static struct connman_network_driver pan_driver = {
 	.disconnect	= pan_disconnect,
 };
 
-static int bluetooth_probe(struct connman_device *device)
+static gboolean network_changed(DBusConnection *connection,
+				DBusMessage *message, void *user_data)
 {
-	DBG("device %p", device);
+	const char *path = dbus_message_get_path(message);
+	struct connman_network *network;
+	DBusMessageIter iter, value;
+	const char *key;
 
-	return 0;
-}
+	DBG("path %s", path);
 
-static void bluetooth_remove(struct connman_device *device)
-{
-	DBG("device %p", device);
-}
+	network = g_hash_table_lookup(bluetooth_networks, path);
+	if (network == NULL)
+		return TRUE;
 
-static void powered_reply(DBusPendingCall *call, void *user_data)
-{
-	DBusMessage *reply;
+	if (dbus_message_iter_init(message, &iter) == FALSE)
+		return TRUE;
 
-	DBG("");
+	dbus_message_iter_get_basic(&iter, &key);
 
-	reply = dbus_pending_call_steal_reply(call);
+	dbus_message_iter_next(&iter);
+	dbus_message_iter_recurse(&iter, &value);
 
-	dbus_message_unref(reply);
-}
+	if (g_str_equal(key, "Connected") == TRUE) {
+		dbus_bool_t connected;
 
-static int change_powered(DBusConnection *connection, const char *path,
-							dbus_bool_t powered)
-{
-	DBusMessage *message;
-	DBusMessageIter iter;
-	DBusPendingCall *call;
+		dbus_message_iter_get_basic(&value, &connected);
 
-	DBG("");
+		if (connected == TRUE)
+			return TRUE;
 
-	if (path == NULL)
-		return -EINVAL;
-
-	message = dbus_message_new_method_call(BLUEZ_SERVICE, path,
-					BLUEZ_ADAPTER_INTERFACE, SET_PROPERTY);
-	if (message == NULL)
-		return -ENOMEM;
-
-	dbus_message_set_auto_start(message, FALSE);
-
-	dbus_message_iter_init_append(message, &iter);
-	connman_dbus_property_append_variant(&iter, "Powered",
-						DBUS_TYPE_BOOLEAN, &powered);
-
-	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
-		connman_error("Failed to change Powered property");
-		dbus_message_unref(message);
-		return -EINVAL;
+		connman_network_set_associating(network, FALSE);
+		connman_network_set_connected(network, FALSE);
 	}
 
-	if (call == NULL) {
-		connman_error("D-Bus connection not available");
-		dbus_message_unref(message);
-		return -EINVAL;
-	}
-
-	dbus_pending_call_set_notify(call, powered_reply, NULL, NULL);
-
-	dbus_message_unref(message);
-
-	return -EINPROGRESS;
+	return TRUE;
 }
 
-static int bluetooth_enable(struct connman_device *device)
-{
-	const char *path = connman_device_get_string(device, "Path");
-
-	DBG("device %p", device);
-
-	return change_powered(connection, path, TRUE);
-}
-
-static int bluetooth_disable(struct connman_device *device)
-{
-	const char *path = connman_device_get_string(device, "Path");
-
-	DBG("device %p", device);
-
-	return change_powered(connection, path, FALSE);
-}
-
-static struct connman_device_driver bluetooth_driver = {
-	.name		= "bluetooth",
-	.type		= CONNMAN_DEVICE_TYPE_BLUETOOTH,
-	.probe		= bluetooth_probe,
-	.remove		= bluetooth_remove,
-	.enable		= bluetooth_enable,
-	.disable	= bluetooth_disable,
-};
-
 static void extract_properties(DBusMessage *reply, const char **parent,
 						const char **address,
 						const char **name,
@@ -430,8 +388,6 @@ static void network_properties_reply(DBusPendingCall *call, void *user_data)
 	char ident[13];
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	extract_properties(reply, &parent, &address, NULL, &name,
 						NULL, NULL, &uuids, NULL);
@@ -478,8 +434,12 @@ static void network_properties_reply(DBusPendingCall *call, void *user_data)
 
 	connman_network_set_group(network, ident);
 
+	g_hash_table_insert(bluetooth_networks, g_strdup(path), network);
+
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static void add_network(struct connman_device *device, const char *path)
@@ -535,7 +495,8 @@ static void check_networks(struct connman_device *device,
 	}
 }
 
-static void adapter_changed(DBusConnection *connection, DBusMessage *message)
+static gboolean adapter_changed(DBusConnection *connection,
+				DBusMessage *message, void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
 	struct connman_device *device;
@@ -546,10 +507,10 @@ static void adapter_changed(DBusConnection *connection, DBusMessage *message)
 
 	device = g_hash_table_lookup(bluetooth_devices, path);
 	if (device == NULL)
-		return;
+		return TRUE;
 
 	if (dbus_message_iter_init(message, &iter) == FALSE)
-		return;
+		return TRUE;
 
 	dbus_message_iter_get_basic(&iter, &key);
 
@@ -569,6 +530,8 @@ static void adapter_changed(DBusConnection *connection, DBusMessage *message)
 	} else if (g_str_equal(key, "Devices") == TRUE) {
 		check_networks(device, &value);
 	}
+
+	return TRUE;
 }
 
 static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
@@ -585,8 +548,6 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 	DBG("path %s", path);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (path == NULL)
 		goto done;
@@ -619,6 +580,8 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 
 	connman_device_set_mode(device, CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE);
 
+	connman_device_set_string(device, "Path", path);
+
 	if (connman_device_register(device) < 0) {
 		connman_device_unref(device);
 		goto done;
@@ -629,6 +592,7 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 update:
 	connman_device_set_string(device, "Address", address);
 	connman_device_set_string(device, "Name", name);
+	connman_device_set_string(device, "Path", path);
 
 	connman_device_set_powered(device, powered);
 	connman_device_set_scanning(device, scanning);
@@ -638,6 +602,8 @@ update:
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static void add_adapter(DBusConnection *connection, const char *path)
@@ -672,6 +638,17 @@ done:
 	dbus_message_unref(message);
 }
 
+static gboolean adapter_added(DBusConnection *connection, DBusMessage *message,
+				void *user_data)
+{
+	const char *path;
+
+	dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+				DBUS_TYPE_INVALID);
+	add_adapter(connection, path);
+	return TRUE;
+}
+
 static void remove_adapter(DBusConnection *connection, const char *path)
 {
 	DBG("path %s", path);
@@ -679,6 +656,17 @@ static void remove_adapter(DBusConnection *connection, const char *path)
 	g_hash_table_remove(bluetooth_devices, path);
 }
 
+static gboolean adapter_removed(DBusConnection *connection, DBusMessage *message,
+				void *user_data)
+{
+	const char *path;
+
+	dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+				DBUS_TYPE_INVALID);
+	remove_adapter(connection, path);
+	return TRUE;
+}
+
 static void list_adapters_reply(DBusPendingCall *call, void *user_data)
 {
 	DBusMessage *reply;
@@ -689,11 +677,15 @@ static void list_adapters_reply(DBusPendingCall *call, void *user_data)
 	DBG("");
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	dbus_error_init(&error);
 
+	if (dbus_set_error_from_message(&error, reply) == TRUE) {
+		connman_error("%s", error.message);
+		dbus_error_free(&error);
+		goto done;
+	}
+
 	if (dbus_message_get_args(reply, &error,
 				DBUS_TYPE_ARRAY, DBUS_TYPE_OBJECT_PATH,
 						&adapters, &num_adapters,
@@ -713,6 +705,8 @@ static void list_adapters_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static void unregister_device(gpointer data)
@@ -735,6 +729,9 @@ static void bluetooth_connect(DBusConnection *connection, void *user_data)
 	bluetooth_devices = g_hash_table_new_full(g_str_hash, g_str_equal,
 						g_free, unregister_device);
 
+	bluetooth_networks = g_hash_table_new_full(g_str_hash, g_str_equal,
+						g_free, NULL);
+
 	message = dbus_message_new_method_call(BLUEZ_SERVICE, "/",
 				BLUEZ_MANAGER_INTERFACE, LIST_ADAPTERS);
 	if (message == NULL)
@@ -766,56 +763,184 @@ static void bluetooth_disconnect(DBusConnection *connection, void *user_data)
 	if (bluetooth_devices == NULL)
 		return;
 
+	g_hash_table_destroy(bluetooth_networks);
 	g_hash_table_destroy(bluetooth_devices);
 	bluetooth_devices = NULL;
 }
 
-static DBusHandlerResult bluetooth_signal(DBusConnection *connection,
-					DBusMessage *message, void *user_data)
+static int bluetooth_probe(struct connman_device *device)
 {
-	if (dbus_message_is_signal(message, BLUEZ_ADAPTER_INTERFACE,
-						PROPERTY_CHANGED) == TRUE) {
-		adapter_changed(connection, message);
-	} else if (dbus_message_is_signal(message, BLUEZ_MANAGER_INTERFACE,
-						ADAPTER_ADDED) == TRUE) {
-		const char *path;
-		dbus_message_get_args(message, NULL,
-					DBUS_TYPE_OBJECT_PATH, &path,
-							DBUS_TYPE_INVALID);
-		add_adapter(connection, path);
-	} else if (dbus_message_is_signal(message, BLUEZ_MANAGER_INTERFACE,
-						ADAPTER_REMOVED) == TRUE) {
-		const char *path;
-		dbus_message_get_args(message, NULL,
-					DBUS_TYPE_OBJECT_PATH, &path,
-							DBUS_TYPE_INVALID);
-		remove_adapter(connection, path);
-	}
+	DBG("device %p", device);
 
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	return 0;
 }
 
+static void bluetooth_remove(struct connman_device *device)
+{
+	DBG("device %p", device);
+}
+
+static void powered_reply(DBusPendingCall *call, void *user_data)
+{
+	DBusError error;
+	DBusMessage *reply;
+
+	DBG("");
+
+	reply = dbus_pending_call_steal_reply(call);
+
+	dbus_error_init(&error);
+
+	if (dbus_set_error_from_message(&error, reply) == TRUE) {
+		connman_error("%s", error.message);
+		dbus_error_free(&error);
+		dbus_message_unref(reply);
+		dbus_pending_call_unref(call);
+		return;
+	}
+
+	dbus_message_unref(reply);
+	dbus_pending_call_unref(call);
+
+	add_adapter(connection, user_data);
+}
+
+static int change_powered(DBusConnection *connection, const char *path,
+							dbus_bool_t powered)
+{
+	DBusMessage *message;
+	DBusMessageIter iter;
+	DBusPendingCall *call;
+
+	DBG("");
+
+	if (path == NULL)
+		return -EINVAL;
+
+	message = dbus_message_new_method_call(BLUEZ_SERVICE, path,
+					BLUEZ_ADAPTER_INTERFACE, SET_PROPERTY);
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_message_iter_init_append(message, &iter);
+	connman_dbus_property_append_basic(&iter, "Powered",
+						DBUS_TYPE_BOOLEAN, &powered);
+
+	if (dbus_connection_send_with_reply(connection, message,
+						&call, TIMEOUT) == FALSE) {
+		connman_error("Failed to change Powered property");
+		dbus_message_unref(message);
+		return -EINVAL;
+	}
+
+	if (call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		return -EINVAL;
+	}
+
+	dbus_pending_call_set_notify(call, powered_reply,
+					g_strdup(path), g_free);
+
+	dbus_message_unref(message);
+
+	return -EINPROGRESS;
+}
+
+static int bluetooth_enable(struct connman_device *device)
+{
+	const char *path = connman_device_get_string(device, "Path");
+
+	DBG("device %p", device);
+
+	return change_powered(connection, path, TRUE);
+}
+
+static int bluetooth_disable(struct connman_device *device)
+{
+	const char *path = connman_device_get_string(device, "Path");
+
+	DBG("device %p", device);
+
+	return change_powered(connection, path, FALSE);
+}
+
+static struct connman_device_driver bluetooth_driver = {
+	.name		= "bluetooth",
+	.type		= CONNMAN_DEVICE_TYPE_BLUETOOTH,
+	.probe		= bluetooth_probe,
+	.remove		= bluetooth_remove,
+	.enable		= bluetooth_enable,
+	.disable	= bluetooth_disable,
+};
+
+static int tech_probe(struct connman_technology *technology)
+{
+	return 0;
+}
+
+static void tech_remove(struct connman_technology *technology)
+{
+}
+
+static int tech_set_tethering(struct connman_technology *technology,
+						connman_bool_t enabled)
+{
+	return 0;
+}
+
+static struct connman_technology_driver tech_driver = {
+	.name		= "bluetooth",
+	.type		= CONNMAN_SERVICE_TYPE_BLUETOOTH,
+	.probe		= tech_probe,
+	.remove		= tech_remove,
+	.set_tethering	= tech_set_tethering,
+};
+
 static guint watch;
-
-static const char *added_rule = "type=signal,member=" ADAPTER_ADDED
-					",interface=" BLUEZ_MANAGER_INTERFACE;
-static const char *removed_rule = "type=signal,member=" ADAPTER_REMOVED
-					",interface=" BLUEZ_MANAGER_INTERFACE;
-
-static const char *adapter_rule = "type=signal,member=" PROPERTY_CHANGED
-					",interface=" BLUEZ_ADAPTER_INTERFACE;
+static guint added_watch;
+static guint removed_watch;
+static guint adapter_watch;
+static guint network_watch;
 
 static int bluetooth_init(void)
 {
-	int err = -EIO;
+	int err;
 
 	connection = connman_dbus_get_connection();
 	if (connection == NULL)
 		return -EIO;
 
-	if (dbus_connection_add_filter(connection, bluetooth_signal,
-							NULL, NULL) == FALSE)
-		goto unref;
+	watch = g_dbus_add_service_watch(connection, BLUEZ_SERVICE,
+			bluetooth_connect, bluetooth_disconnect, NULL, NULL);
+
+	added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_MANAGER_INTERFACE,
+						ADAPTER_ADDED, adapter_added,
+						NULL, NULL);
+
+	removed_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_MANAGER_INTERFACE,
+						ADAPTER_REMOVED, adapter_removed,
+						NULL, NULL);
+
+	adapter_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_ADAPTER_INTERFACE,
+						PROPERTY_CHANGED, adapter_changed,
+						NULL, NULL);
+
+	network_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_NETWORK_INTERFACE,
+						PROPERTY_CHANGED, network_changed,
+						NULL, NULL);
+
+	if (watch == 0 || added_watch == 0 || removed_watch == 0
+			|| adapter_watch == 0 || network_watch == 0) {
+		err = -EIO;
+		goto remove;
+	}
 
 	err = connman_network_driver_register(&pan_driver);
 	if (err < 0)
@@ -827,26 +952,22 @@ static int bluetooth_init(void)
 		goto remove;
 	}
 
-	watch = g_dbus_add_service_watch(connection, BLUEZ_SERVICE,
-			bluetooth_connect, bluetooth_disconnect, NULL, NULL);
-	if (watch == 0) {
+	err = connman_technology_driver_register(&tech_driver);
+	if (err < 0) {
 		connman_device_driver_unregister(&bluetooth_driver);
 		connman_network_driver_unregister(&pan_driver);
-		err = -EIO;
-		goto remove;
+		return err;
 	}
 
-	dbus_bus_add_match(connection, added_rule, NULL);
-	dbus_bus_add_match(connection, removed_rule, NULL);
-	dbus_bus_add_match(connection, adapter_rule, NULL);
-	dbus_connection_flush(connection);
-
 	return 0;
 
 remove:
-	dbus_connection_remove_filter(connection, bluetooth_signal, NULL);
+	g_dbus_remove_watch(connection, watch);
+	g_dbus_remove_watch(connection, added_watch);
+	g_dbus_remove_watch(connection, removed_watch);
+	g_dbus_remove_watch(connection, adapter_watch);
+	g_dbus_remove_watch(connection, network_watch);
 
-unref:
 	dbus_connection_unref(connection);
 
 	return err;
@@ -854,20 +975,19 @@ unref:
 
 static void bluetooth_exit(void)
 {
-	dbus_bus_remove_match(connection, adapter_rule, NULL);
-	dbus_bus_remove_match(connection, removed_rule, NULL);
-	dbus_bus_remove_match(connection, added_rule, NULL);
-	dbus_connection_flush(connection);
-
 	g_dbus_remove_watch(connection, watch);
+	g_dbus_remove_watch(connection, added_watch);
+	g_dbus_remove_watch(connection, removed_watch);
+	g_dbus_remove_watch(connection, adapter_watch);
+	g_dbus_remove_watch(connection, network_watch);
 
 	bluetooth_disconnect(connection, NULL);
 
+	connman_technology_driver_unregister(&tech_driver);
+
 	connman_device_driver_unregister(&bluetooth_driver);
 	connman_network_driver_unregister(&pan_driver);
 
-	dbus_connection_remove_filter(connection, bluetooth_signal, NULL);
-
 	dbus_connection_unref(connection);
 }
 
