@@ -56,8 +56,14 @@ struct connman_ipdevice {
 	unsigned int flags;
 	char *address;
 	uint16_t mtu;
-	uint32_t tx_bytes;
+	uint32_t rx_packets;
+	uint32_t tx_packets;
 	uint32_t rx_bytes;
+	uint32_t tx_bytes;
+	uint32_t rx_errors;
+	uint32_t tx_errors;
+	uint32_t rx_dropped;
+	uint32_t tx_dropped;
 
 	GSList *address_list;
 	char *gateway;
@@ -364,11 +370,20 @@ static void update_stats(struct connman_ipdevice *ipdevice,
 	if (ipdevice->config == NULL)
 		return;
 
+	ipdevice->rx_packets = stats->rx_packets;
+	ipdevice->tx_packets = stats->tx_packets;
 	ipdevice->rx_bytes = stats->rx_bytes;
 	ipdevice->tx_bytes = stats->tx_bytes;
+	ipdevice->rx_errors = stats->rx_errors;
+	ipdevice->tx_errors = stats->tx_errors;
+	ipdevice->rx_dropped = stats->rx_dropped;
+	ipdevice->tx_dropped = stats->tx_dropped;
 
-	__connman_counter_notify(ipdevice->ifname,
-				ipdevice->rx_bytes, ipdevice->tx_bytes);
+	__connman_counter_notify(ipdevice->config,
+				ipdevice->rx_packets, ipdevice->tx_packets,
+				ipdevice->rx_bytes, ipdevice->tx_bytes,
+				ipdevice->rx_errors, ipdevice->tx_errors,
+				ipdevice->rx_dropped, ipdevice->tx_dropped);
 }
 
 void __connman_ipconfig_newlink(int index, unsigned short type,
@@ -381,7 +396,6 @@ void __connman_ipconfig_newlink(int index, unsigned short type,
 	GString *str;
 	gboolean up = FALSE, down = FALSE;
 	gboolean lower_up = FALSE, lower_down = FALSE;
-	char *ifname;
 
 	DBG("index %d", index);
 
@@ -392,22 +406,12 @@ void __connman_ipconfig_newlink(int index, unsigned short type,
 	if (ipdevice != NULL)
 		goto update;
 
-	ifname = connman_inet_ifname(index);
-
-	if (__connman_element_device_isfiltered(ifname) == TRUE) {
-		connman_info("Ignoring interface %s (filtered)", ifname);
-		g_free(ifname);
-		return;
-	}
-
 	ipdevice = g_try_new0(struct connman_ipdevice, 1);
-	if (ipdevice == NULL) {
-		g_free(ifname);
+	if (ipdevice == NULL)
 		return;
-	}
 
 	ipdevice->index = index;
-	ipdevice->ifname = ifname;
+	ipdevice->ifname = connman_inet_ifname(index);
 	ipdevice->type = type;
 
 	ipdevice->address = g_strdup(address);
@@ -634,8 +638,23 @@ void __connman_ipconfig_newroute(int index, unsigned char scope,
 		return;
 
 	if (scope == 0 && g_strcmp0(dst, "0.0.0.0") == 0) {
+		GSList *list;
+
 		g_free(ipdevice->gateway);
 		ipdevice->gateway = g_strdup(gateway);
+
+		if (ipdevice->config != NULL &&
+					ipdevice->config->system != NULL) {
+			g_free(ipdevice->config->system->gateway);
+			ipdevice->config->system->gateway = g_strdup(gateway);
+		}
+
+		for (list = ipdevice->address_list; list; list = list->next) {
+			struct connman_ipaddress *ipaddress = list->data;
+
+			g_free(ipaddress->gateway);
+			ipaddress->gateway = g_strdup(gateway);
+		}
 	}
 
 	connman_info("%s {add} route %s gw %s scope %u <%s>",
@@ -655,8 +674,23 @@ void __connman_ipconfig_delroute(int index, unsigned char scope,
 		return;
 
 	if (scope == 0 && g_strcmp0(dst, "0.0.0.0") == 0) {
+		GSList *list;
+
 		g_free(ipdevice->gateway);
 		ipdevice->gateway = NULL;
+
+		if (ipdevice->config != NULL &&
+					ipdevice->config->system != NULL) {
+			g_free(ipdevice->config->system->gateway);
+			ipdevice->config->system->gateway = NULL;
+		}
+
+		for (list = ipdevice->address_list; list; list = list->next) {
+			struct connman_ipaddress *ipaddress = list->data;
+
+			g_free(ipaddress->gateway);
+			ipaddress->gateway = NULL;
+		}
 	}
 
 	connman_info("%s {del} route %s gw %s scope %u <%s>",
@@ -712,7 +746,14 @@ const char *__connman_ipconfig_get_gateway(int index)
 	if (ipdevice == NULL)
 		return NULL;
 
-	return ipdevice->gateway;
+	if (ipdevice->gateway != NULL)
+		return ipdevice->gateway;
+
+	if (ipdevice->config != NULL &&
+			ipdevice->config->address != NULL)
+		return ipdevice->config->address->gateway;
+
+	return NULL;
 }
 
 void __connman_ipconfig_set_index(struct connman_ipconfig *ipconfig, int index)
@@ -848,6 +889,9 @@ void connman_ipconfig_set_data(struct connman_ipconfig *ipconfig, void *data)
  */
 int connman_ipconfig_get_index(struct connman_ipconfig *ipconfig)
 {
+	if (ipconfig == NULL)
+		return -1;
+
 	if (ipconfig->origin != NULL)
 		return ipconfig->origin->index;
 
@@ -864,6 +908,9 @@ const char *connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
 {
 	struct connman_ipdevice *ipdevice;
 
+	if (ipconfig == NULL)
+		return NULL;
+
 	if (ipconfig->index < 0)
 		return NULL;
 
@@ -903,8 +950,7 @@ int connman_ipconfig_set_method(struct connman_ipconfig *ipconfig,
 	return 0;
 }
 
-enum connman_ipconfig_method __connman_ipconfig_get_method(
-				struct connman_ipconfig *ipconfig)
+enum connman_ipconfig_method __connman_ipconfig_get_method(struct connman_ipconfig *ipconfig)
 {
 	if (ipconfig == NULL)
 		return CONNMAN_IPCONFIG_METHOD_UNKNOWN;
@@ -993,6 +1039,8 @@ int __connman_ipconfig_clear_address(struct connman_ipconfig *ipconfig)
 int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 {
 	struct connman_ipdevice *ipdevice;
+	gboolean up = FALSE, down = FALSE;
+	gboolean lower_up = FALSE, lower_down = FALSE;
 
 	DBG("ipconfig %p", ipconfig);
 
@@ -1019,6 +1067,27 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 
 	ipconfig_list = g_list_append(ipconfig_list, ipconfig);
 
+	if (ipdevice->flags & IFF_UP)
+		up = TRUE;
+	else
+		down = TRUE;
+
+	if ((ipdevice->flags & (IFF_RUNNING | IFF_LOWER_UP)) ==
+			(IFF_RUNNING | IFF_LOWER_UP))
+		lower_up = TRUE;
+	else if ((ipdevice->flags & (IFF_RUNNING | IFF_LOWER_UP)) == 0)
+		lower_down = TRUE;
+
+	if (up == TRUE && ipconfig->ops->up)
+		ipconfig->ops->up(ipconfig);
+	if (lower_up == TRUE && ipconfig->ops->lower_up)
+		ipconfig->ops->lower_up(ipconfig);
+
+	if (lower_down == TRUE && ipconfig->ops->lower_down)
+		ipconfig->ops->lower_down(ipconfig);
+	if (down == TRUE && ipconfig->ops->down)
+		ipconfig->ops->down(ipconfig);
+
 	return 0;
 }
 
@@ -1109,6 +1178,10 @@ void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
 		connman_dbus_dict_append_basic(iter, "Netmask",
 						DBUS_TYPE_STRING, &mask);
 	}
+
+	if (ipconfig->system->gateway != NULL)
+		connman_dbus_dict_append_basic(iter, "Gateway",
+				DBUS_TYPE_STRING, &ipconfig->system->gateway);
 }
 
 void __connman_ipconfig_append_ipv4config(struct connman_ipconfig *ipconfig,
@@ -1149,6 +1222,10 @@ void __connman_ipconfig_append_ipv4config(struct connman_ipconfig *ipconfig,
 		connman_dbus_dict_append_basic(iter, "Netmask",
 						DBUS_TYPE_STRING, &mask);
 	}
+
+	if (ipconfig->address->gateway != NULL)
+		connman_dbus_dict_append_basic(iter, "Gateway",
+				DBUS_TYPE_STRING, &ipconfig->address->gateway);
 }
 
 int __connman_ipconfig_set_ipv4config(struct connman_ipconfig *ipconfig,
