@@ -219,6 +219,7 @@ int main(int argc, char *argv[])
 	__connman_element_init(option_device, option_nodevice);
 
 	__connman_agent_init();
+	__connman_tethering_init();
 	__connman_counter_init();
 	__connman_manager_init(option_compat);
 	__connman_profile_init();
@@ -230,6 +231,7 @@ int main(int argc, char *argv[])
 	__connman_udev_init();
 	__connman_task_init();
 	__connman_session_init();
+	__connman_timeserver_init();
 
 	__connman_plugin_init(option_plugin, option_noplugin);
 
@@ -251,6 +253,7 @@ int main(int argc, char *argv[])
 
 	__connman_plugin_cleanup();
 
+	__connman_timeserver_cleanup();
 	__connman_session_cleanup();
 	__connman_task_cleanup();
 	__connman_udev_cleanup();
@@ -263,6 +266,7 @@ int main(int argc, char *argv[])
 	__connman_manager_cleanup();
 	__connman_counter_cleanup();
 	__connman_agent_cleanup();
+	__connman_tethering_cleanup();
 
 	__connman_element_cleanup();
 	__connman_storage_cleanup();
