@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -27,180 +27,12 @@
 
 #include "connman.h"
 
-static void append_profiles(DBusMessageIter *dict)
-{
-	DBusMessageIter entry, value, iter;
-	const char *key = "Profiles";
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-		DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_OBJECT_PATH_AS_STRING,
-								&value);
-
-	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
-				DBUS_TYPE_OBJECT_PATH_AS_STRING, &iter);
-	__connman_profile_list(&iter);
-	dbus_message_iter_close_container(&value, &iter);
-
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
-}
-
-static void append_services(DBusMessageIter *dict)
-{
-	DBusMessageIter entry, value, iter;
-	const char *key = "Services";
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-		DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_OBJECT_PATH_AS_STRING,
-								&value);
-
-	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
-				DBUS_TYPE_OBJECT_PATH_AS_STRING, &iter);
-	__connman_service_list(&iter);
-	dbus_message_iter_close_container(&value, &iter);
-
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
-}
-
-static void append_devices(DBusMessageIter *dict)
-{
-	DBusMessageIter entry, value, iter;
-	const char *key = "Devices";
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-		DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_OBJECT_PATH_AS_STRING,
-								&value);
-
-	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
-				DBUS_TYPE_OBJECT_PATH_AS_STRING, &iter);
-	__connman_element_list(NULL, CONNMAN_ELEMENT_TYPE_DEVICE, &iter);
-	dbus_message_iter_close_container(&value, &iter);
-
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
-}
-
-static void append_connections(DBusMessageIter *dict)
-{
-	DBusMessageIter entry, value, iter;
-	const char *key = "Connections";
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-		DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_OBJECT_PATH_AS_STRING,
-								&value);
-
-	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
-				DBUS_TYPE_OBJECT_PATH_AS_STRING, &iter);
-	__connman_element_list(NULL, CONNMAN_ELEMENT_TYPE_CONNECTION, &iter);
-	dbus_message_iter_close_container(&value, &iter);
-
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
-}
-
-static void append_available_technologies(DBusMessageIter *dict)
-{
-	DBusMessageIter entry, value, iter;
-	const char *key = "AvailableTechnologies";
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-			DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_STRING_AS_STRING,
-								&value);
-
-	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
-					DBUS_TYPE_STRING_AS_STRING, &iter);
-	__connman_notifier_list_registered(&iter);
-	dbus_message_iter_close_container(&value, &iter);
-
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
-}
-
-static void append_enabled_technologies(DBusMessageIter *dict)
-{
-	DBusMessageIter entry, value, iter;
-	const char *key = "EnabledTechnologies";
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-			DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_STRING_AS_STRING,
-								&value);
-
-	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
-					DBUS_TYPE_STRING_AS_STRING, &iter);
-	__connman_notifier_list_enabled(&iter);
-	dbus_message_iter_close_container(&value, &iter);
-
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
-}
-
-static void append_connected_technologies(DBusMessageIter *dict)
-{
-	DBusMessageIter entry, value, iter;
-	const char *key = "ConnectedTechnologies";
-
-	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
-								NULL, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-			DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_STRING_AS_STRING,
-								&value);
-
-	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
-					DBUS_TYPE_STRING_AS_STRING, &iter);
-	__connman_notifier_list_connected(&iter);
-	dbus_message_iter_close_container(&value, &iter);
-
-	dbus_message_iter_close_container(&entry, &value);
-
-	dbus_message_iter_close_container(dict, &entry);
-}
-
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
-	connman_bool_t offlinemode;
+	connman_bool_t offlinemode, tethering;
 	const char *str;
 
 	DBG("conn %p", conn);
@@ -215,44 +47,52 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	dbus_message_iter_init_append(reply, &array);
 
-	dbus_message_iter_open_container(&array, DBUS_TYPE_ARRAY,
-			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
-			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
-			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+	connman_dbus_dict_open(&array, &dict);
 
 	str = __connman_profile_active_path();
 	if (str != NULL)
-		connman_dbus_dict_append_variant(&dict, "ActiveProfile",
+		connman_dbus_dict_append_basic(&dict, "ActiveProfile",
 						DBUS_TYPE_OBJECT_PATH, &str);
 
-	append_profiles(&dict);
-	append_services(&dict);
+	connman_dbus_dict_append_array(&dict, "Profiles",
+			DBUS_TYPE_OBJECT_PATH, __connman_profile_list, NULL);
+	connman_dbus_dict_append_array(&dict, "Services",
+			DBUS_TYPE_OBJECT_PATH, __connman_service_list, NULL);
+	connman_dbus_dict_append_array(&dict, "Providers",
+			DBUS_TYPE_OBJECT_PATH, __connman_provider_list, NULL);
+	connman_dbus_dict_append_array(&dict, "Technologies",
+			DBUS_TYPE_OBJECT_PATH, __connman_technology_list, NULL);
 
-	append_devices(&dict);
-	append_connections(&dict);
-
-	if (__connman_element_count(NULL, CONNMAN_ELEMENT_TYPE_CONNECTION) > 0)
-		str = "online";
-	else
-		str = "offline";
-
-	connman_dbus_dict_append_variant(&dict, "State",
+	str = __connman_notifier_get_state();
+	connman_dbus_dict_append_basic(&dict, "State",
 						DBUS_TYPE_STRING, &str);
 
 	offlinemode = __connman_profile_get_offlinemode();
-	connman_dbus_dict_append_variant(&dict, "OfflineMode",
+	connman_dbus_dict_append_basic(&dict, "OfflineMode",
 					DBUS_TYPE_BOOLEAN, &offlinemode);
 
-	append_available_technologies(&dict);
-	append_enabled_technologies(&dict);
-	append_connected_technologies(&dict);
+	tethering = __connman_tethering_get_status();
+	connman_dbus_dict_append_basic(&dict, "Tethering",
+					DBUS_TYPE_BOOLEAN, &tethering);
+
+	connman_dbus_dict_append_array(&dict, "AvailableTechnologies",
+		DBUS_TYPE_STRING, __connman_notifier_list_registered, NULL);
+	connman_dbus_dict_append_array(&dict, "EnabledTechnologies",
+		DBUS_TYPE_STRING, __connman_notifier_list_enabled, NULL);
+	connman_dbus_dict_append_array(&dict, "ConnectedTechnologies",
+		DBUS_TYPE_STRING, __connman_notifier_list_connected, NULL);
 
 	str = __connman_service_default();
 	if (str != NULL)
-		connman_dbus_dict_append_variant(&dict, "DefaultTechnology",
+		connman_dbus_dict_append_basic(&dict, "DefaultTechnology",
 						DBUS_TYPE_STRING, &str);
 
-	dbus_message_iter_close_container(&array, &dict);
+	connman_dbus_dict_append_array(&dict, "AvailableDebugs",
+			DBUS_TYPE_STRING, __connman_debug_list_available, NULL);
+	connman_dbus_dict_append_array(&dict, "EnabledDebugs",
+			DBUS_TYPE_STRING, __connman_debug_list_enabled, NULL);
+
+	connman_dbus_dict_close(&array, &dict);
 
 	return reply;
 }
@@ -290,6 +130,20 @@ static DBusMessage *set_property(DBusConnection *conn,
 		__connman_profile_set_offlinemode(offlinemode);
 
 		__connman_profile_save_default();
+	} else if (g_str_equal(name, "Tethering") == TRUE) {
+		connman_bool_t tethering;
+
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __connman_error_invalid_arguments(msg);
+
+		dbus_message_iter_get_basic(&value, &tethering);
+
+		if (__connman_tethering_set_status(tethering) < 0)
+			return __connman_error_invalid_arguments(msg);
+
+		connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
+					CONNMAN_MANAGER_INTERFACE, "Tethering",
+						DBUS_TYPE_BOOLEAN, &tethering);
 	} else if (g_str_equal(name, "ActiveProfile") == TRUE) {
 		const char *str;
 
@@ -313,10 +167,7 @@ static DBusMessage *get_state(DBusConnection *conn,
 					CONNMAN_SECURITY_PRIVILEGE_PUBLIC) < 0)
 		return __connman_error_permission_denied(msg);
 
-	if (__connman_element_count(NULL, CONNMAN_ELEMENT_TYPE_CONNECTION) > 0)
-		str = "online";
-	else
-		str = "offline";
+	str = __connman_notifier_get_state();
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &str,
 						DBUS_TYPE_INVALID);
@@ -367,6 +218,28 @@ static DBusMessage *remove_profile(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static DBusMessage *remove_provider(DBusConnection *conn,
+				    DBusMessage *msg, void *data)
+{
+	const char *path;
+	int err;
+
+	DBG("conn %p", conn);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &path,
+			      DBUS_TYPE_INVALID);
+
+	if (__connman_security_check_privilege(msg,
+				CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
+		return __connman_error_permission_denied(msg);
+
+	err = __connman_provider_remove(path);
+	if (err < 0)
+		return __connman_error_failed(msg, -err);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
 static DBusMessage *request_scan(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -490,6 +363,9 @@ static DBusMessage *enable_technology(DBusConnection *conn,
 	else
 		return __connman_error_invalid_arguments(msg);
 
+	if (__connman_notifier_is_registered(type) == FALSE)
+		return __connman_error_not_registered(msg);
+
 	if (__connman_notifier_is_enabled(type) == TRUE)
 		return __connman_error_already_enabled(msg);
 
@@ -535,6 +411,9 @@ static DBusMessage *disable_technology(DBusConnection *conn,
 	else
 		return __connman_error_invalid_arguments(msg);
 
+	if (__connman_notifier_is_registered(type) == FALSE)
+		return __connman_error_not_registered(msg);
+
 	if (__connman_notifier_is_enabled(type) == FALSE)
 		return __connman_error_already_disabled(msg);
 
@@ -552,6 +431,54 @@ static DBusMessage *disable_technology(DBusConnection *conn,
 	return NULL;
 }
 
+static DBusMessage *get_services(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	DBusMessage *reply;
+	DBusMessageIter iter, array;
+
+	reply = dbus_message_new_method_return(msg);
+	if (reply == NULL)
+		return NULL;
+
+	dbus_message_iter_init_append(reply, &iter);
+
+	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
+			DBUS_STRUCT_BEGIN_CHAR_AS_STRING
+			DBUS_TYPE_OBJECT_PATH_AS_STRING
+			DBUS_TYPE_ARRAY_AS_STRING
+				DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+					DBUS_TYPE_STRING_AS_STRING
+					DBUS_TYPE_VARIANT_AS_STRING
+				DBUS_DICT_ENTRY_END_CHAR_AS_STRING
+			DBUS_STRUCT_END_CHAR_AS_STRING, &array);
+
+	__connman_service_list_struct(&array);
+
+	dbus_message_iter_close_container(&iter, &array);
+
+	return reply;
+}
+
+static DBusMessage *lookup_service(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *pattern, *path;
+	int err;
+
+	DBG("conn %p", conn);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &pattern,
+							DBUS_TYPE_INVALID);
+
+	err = __connman_service_lookup(pattern, &path);
+	if (err < 0)
+		return __connman_error_failed(msg, -err);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID);
+}
+
 static DBusMessage *connect_service(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -576,6 +503,31 @@ static DBusMessage *connect_service(DBusConnection *conn,
 	return NULL;
 }
 
+
+static DBusMessage *connect_provider(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	int err;
+
+	DBG("conn %p", conn);
+
+	if (__connman_security_check_privilege(msg,
+				CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
+		return __connman_error_permission_denied(msg);
+
+	err = __connman_provider_create_and_connect(msg);
+	if (err < 0) {
+		if (err == -EINPROGRESS) {
+			connman_error("Invalid return code from connect");
+			err = -EINVAL;
+		}
+
+		return __connman_error_failed(msg, -err);
+	}
+
+	return NULL;
+}
+
 static DBusMessage *register_agent(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -616,21 +568,115 @@ static DBusMessage *unregister_agent(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static DBusMessage *register_counter(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *sender, *path;
+	unsigned int accuracy, period;
+	int err;
+
+	DBG("conn %p", conn);
+
+	sender = dbus_message_get_sender(msg);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+						DBUS_TYPE_UINT32, &accuracy,
+						DBUS_TYPE_UINT32, &period,
+							DBUS_TYPE_INVALID);
+
+	/* FIXME: add handling of accuracy parameter */
+
+	err = __connman_counter_register(sender, path, period);
+	if (err < 0)
+		return __connman_error_failed(msg, -err);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *unregister_counter(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *sender, *path;
+	int err;
+
+	DBG("conn %p", conn);
+
+	sender = dbus_message_get_sender(msg);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID);
+
+	err = __connman_counter_unregister(sender, path);
+	if (err < 0)
+		return __connman_error_failed(msg, -err);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *request_session(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *bearer, *sender, *service_path;
+	struct connman_service *service;
+
+	DBG("conn %p", conn);
+
+	sender = dbus_message_get_sender(msg);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &bearer,
+							DBUS_TYPE_INVALID);
+
+	service = __connman_session_request(bearer, sender);
+	if (service == NULL)
+		return __connman_error_failed(msg, EINVAL);
+
+	service_path = __connman_service_get_path(service);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_OBJECT_PATH, &service_path,
+						DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *release_session(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *sender;
+	int err;
+
+	DBG("conn %p", conn);
+
+	sender = dbus_message_get_sender(msg);
+
+	err = __connman_session_release(sender);
+	if (err < 0)
+		return __connman_error_failed(msg, -err);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
 static GDBusMethodTable manager_methods[] = {
 	{ "GetProperties",     "",      "a{sv}", get_properties     },
 	{ "SetProperty",       "sv",    "",      set_property       },
 	{ "GetState",          "",      "s",     get_state          },
 	{ "CreateProfile",     "s",     "o",     create_profile     },
 	{ "RemoveProfile",     "o",     "",      remove_profile     },
+	{ "RemoveProvider",    "s",     "",      remove_provider    },
 	{ "RequestScan",       "s",     "",      request_scan       },
 	{ "EnableTechnology",  "s",     "",      enable_technology,
 						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "DisableTechnology", "s",     "",      disable_technology,
 						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "GetServices",       "",      "a(oa{sv})", get_services   },
+	{ "LookupService",     "s",     "o",     lookup_service,    },
 	{ "ConnectService",    "a{sv}", "o",     connect_service,
 						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "ConnectProvider",   "a{sv}", "o",     connect_provider,
+						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "RegisterAgent",     "o",     "",      register_agent     },
 	{ "UnregisterAgent",   "o",     "",      unregister_agent   },
+	{ "RegisterCounter",   "ouu",   "",      register_counter   },
+	{ "UnregisterCounter", "o",     "",      unregister_counter },
+	{ "RequestSession",    "s",     "o",     request_session    },
+	{ "ReleaseSession",    "s",     "",      release_session    },
 	{ },
 };
 
@@ -692,7 +738,7 @@ static DBusMessage *nm_state(DBusConnection *conn,
 	if (reply == NULL)
 		return NULL;
 
-	if (__connman_element_count(NULL, CONNMAN_ELEMENT_TYPE_CONNECTION) > 0)
+	if (__connman_notifier_count_connected() > 0)
 		state = NM_STATE_CONNECTED;
 	else
 		state = NM_STATE_DISCONNECTED;
