@@ -36,6 +36,30 @@ static DBusConnection *connection = NULL;
 static GSequence *service_list = NULL;
 static GHashTable *service_hash = NULL;
 
+struct connman_stats {
+	connman_bool_t valid;
+	connman_bool_t enabled;
+	unsigned int rx_packets_last;
+	unsigned int tx_packets_last;
+	unsigned int rx_packets;
+	unsigned int tx_packets;
+	unsigned int rx_bytes_last;
+	unsigned int tx_bytes_last;
+	unsigned int rx_bytes;
+	unsigned int tx_bytes;
+	unsigned int rx_errors_last;
+	unsigned int tx_errors_last;
+	unsigned int rx_errors;
+	unsigned int tx_errors;
+	unsigned int rx_dropped_last;
+	unsigned int tx_dropped_last;
+	unsigned int rx_dropped;
+	unsigned int tx_dropped;
+	unsigned int time_start;
+	unsigned int time;
+	GTimer *timer;
+};
+
 struct connman_service {
 	gint refcount;
 	char *identifier;
@@ -63,6 +87,7 @@ struct connman_service {
 	char *mcc;
 	char *mnc;
 	connman_bool_t roaming;
+	connman_bool_t login_required;
 	struct connman_ipconfig *ipconfig;
 	struct connman_network *network;
 	char **nameservers;
@@ -79,6 +104,7 @@ struct connman_service {
 	DBusMessage *pending;
 	guint timeout;
 	struct connman_location *location;
+	struct connman_stats stats;
 };
 
 static void append_path(gpointer value, gpointer user_data)
@@ -207,8 +233,6 @@ static const char *state2string(enum connman_service_state state)
 		return "configuration";
 	case CONNMAN_SERVICE_STATE_READY:
 		return "ready";
-	case CONNMAN_SERVICE_STATE_LOGIN:
-		return "login";
 	case CONNMAN_SERVICE_STATE_ONLINE:
 		return "online";
 	case CONNMAN_SERVICE_STATE_DISCONNECT:
@@ -256,7 +280,6 @@ static connman_bool_t is_connecting(struct connman_service *service)
 	case CONNMAN_SERVICE_STATE_FAILURE:
 	case CONNMAN_SERVICE_STATE_DISCONNECT:
 	case CONNMAN_SERVICE_STATE_READY:
-	case CONNMAN_SERVICE_STATE_LOGIN:
 	case CONNMAN_SERVICE_STATE_ONLINE:
 		break;
 	case CONNMAN_SERVICE_STATE_ASSOCIATION:
@@ -278,7 +301,6 @@ static connman_bool_t is_connected(const struct connman_service *service)
 	case CONNMAN_SERVICE_STATE_FAILURE:
 		break;
 	case CONNMAN_SERVICE_STATE_READY:
-	case CONNMAN_SERVICE_STATE_LOGIN:
 	case CONNMAN_SERVICE_STATE_ONLINE:
 		return TRUE;
 	}
@@ -304,7 +326,6 @@ static void update_nameservers(struct connman_service *service)
 		connman_resolver_remove_all(ifname);
 		return;
 	case CONNMAN_SERVICE_STATE_READY:
-	case CONNMAN_SERVICE_STATE_LOGIN:
 	case CONNMAN_SERVICE_STATE_ONLINE:
 		break;
 	}
@@ -349,6 +370,214 @@ void __connman_service_remove_nameserver(struct connman_service *service,
 	update_nameservers(service);
 }
 
+void __connman_service_nameserver_add_routes(struct connman_service *service,
+						const char *gw)
+{
+	int index;
+
+	if (service == NULL)
+		return;
+
+	index = connman_network_get_index(service->network);
+
+	if (service->nameservers != NULL) {
+		int i;
+
+		/*
+		 * We add nameservers host routes for nameservers that
+		 * are not on our subnet. For those who are, the subnet
+		 * route will be installed by the time the dns proxy code
+		 * tries to reach them. The subnet route is installed
+		 * when setting the interface IP address.
+		 */
+		for (i = 0; service->nameservers[i]; i++) {
+			if (connman_inet_compare_subnet(index,
+							service->nameservers[i]))
+				continue;
+
+			connman_inet_add_host_route(index,
+						service->nameservers[i], gw);
+		}
+	} else if (service->nameserver != NULL) {
+		if (connman_inet_compare_subnet(index, service->nameserver))
+			return;
+
+		connman_inet_add_host_route(index, service->nameserver, gw);
+	}
+}
+
+void __connman_service_nameserver_del_routes(struct connman_service *service)
+{
+	int index;
+
+	if (service == NULL)
+		return;
+
+	index = connman_network_get_index(service->network);
+
+	if (service->nameservers != NULL) {
+		int i;
+
+		for (i = 0; service->nameservers[i]; i++)
+			connman_inet_del_host_route(index,
+						service->nameservers[i]);
+	} else if (service->nameserver != NULL) {
+		connman_inet_del_host_route(index, service->nameserver);
+	}
+}
+
+static void __connman_service_stats_start(struct connman_service *service)
+{
+	DBG("service %p", service);
+
+	if (service->stats.timer == NULL)
+		return;
+
+	service->stats.enabled = TRUE;
+
+	service->stats.time_start = service->stats.time;
+
+	g_timer_start(service->stats.timer);
+
+	__connman_counter_add_service(service);
+}
+
+static void __connman_service_stats_stop(struct connman_service *service)
+{
+	unsigned int seconds;
+
+	DBG("service %p", service);
+
+	if (service->stats.timer == NULL)
+		return;
+
+	if (service->stats.enabled == FALSE)
+		return;
+
+	__connman_counter_remove_service(service);
+
+	g_timer_stop(service->stats.timer);
+
+	seconds = g_timer_elapsed(service->stats.timer, NULL);
+	service->stats.time = service->stats.time_start + seconds;
+
+	service->stats.enabled = FALSE;
+}
+
+static int __connman_service_stats_load(struct connman_service *service,
+		GKeyFile *keyfile, const char *identifier)
+{
+	service->stats.rx_packets = g_key_file_get_integer(keyfile,
+				identifier, "rx_packets", NULL);
+	service->stats.tx_packets = g_key_file_get_integer(keyfile,
+				identifier, "tx_packets", NULL);
+	service->stats.rx_bytes = g_key_file_get_integer(keyfile,
+				identifier, "rx_bytes", NULL);
+	service->stats.tx_bytes = g_key_file_get_integer(keyfile,
+				identifier, "tx_bytes", NULL);
+	service->stats.rx_errors = g_key_file_get_integer(keyfile,
+				identifier, "rx_errors", NULL);
+	service->stats.tx_errors = g_key_file_get_integer(keyfile,
+				identifier, "tx_errors", NULL);
+	service->stats.rx_dropped = g_key_file_get_integer(keyfile,
+				identifier, "rx_dropped", NULL);
+	service->stats.tx_dropped = g_key_file_get_integer(keyfile,
+				identifier, "tx_dropped", NULL);
+	service->stats.time = g_key_file_get_integer(keyfile,
+				identifier, "time", NULL);
+
+	return 0;
+}
+
+static int __connman_service_stats_save(struct connman_service *service,
+		GKeyFile *keyfile, const char *identifier)
+{
+	g_key_file_set_integer(keyfile, identifier, "rx_packets",
+			service->stats.rx_packets);
+	g_key_file_set_integer(keyfile, identifier, "tx_packets",
+			service->stats.tx_packets);
+	g_key_file_set_integer(keyfile, identifier, "rx_bytes",
+			service->stats.rx_bytes);
+	g_key_file_set_integer(keyfile, identifier, "tx_bytes",
+			service->stats.tx_bytes);
+	g_key_file_set_integer(keyfile, identifier, "rx_errors",
+			service->stats.rx_errors);
+	g_key_file_set_integer(keyfile, identifier, "tx_errors",
+			service->stats.tx_errors);
+	g_key_file_set_integer(keyfile, identifier, "rx_dropped",
+			service->stats.rx_dropped);
+	g_key_file_set_integer(keyfile, identifier, "tx_dropped",
+			service->stats.tx_dropped);
+	g_key_file_set_integer(keyfile, identifier, "time",
+			service->stats.time);
+
+	return 0;
+}
+
+static void __connman_service_reset_stats(struct connman_service *service)
+{
+	DBG("service %p", service);
+
+	service->stats.valid = FALSE;
+	service->stats.rx_packets = 0;
+	service->stats.tx_packets = 0;
+	service->stats.rx_bytes = 0;
+	service->stats.tx_bytes = 0;
+	service->stats.rx_errors = 0;
+	service->stats.tx_errors = 0;
+	service->stats.rx_dropped = 0;
+	service->stats.tx_dropped = 0;
+	service->stats.time = 0;
+	service->stats.time_start = 0;
+	g_timer_reset(service->stats.timer);
+
+}
+
+unsigned long __connman_service_stats_get_rx_packets(struct connman_service *service)
+{
+	return service->stats.rx_packets;
+}
+
+unsigned long __connman_service_stats_get_tx_packets(struct connman_service *service)
+{
+	return service->stats.tx_packets;
+}
+
+unsigned long __connman_service_stats_get_rx_bytes(struct connman_service *service)
+{
+	return service->stats.rx_bytes;
+}
+
+unsigned long __connman_service_stats_get_tx_bytes(struct connman_service *service)
+{
+	return service->stats.tx_bytes;
+}
+
+unsigned long __connman_service_stats_get_rx_errors(struct connman_service *service)
+{
+	return service->stats.rx_errors;
+}
+
+unsigned long __connman_service_stats_get_tx_errors(struct connman_service *service)
+{
+	return service->stats.tx_errors;
+}
+
+unsigned long __connman_service_stats_get_rx_dropped(struct connman_service *service)
+{
+	return service->stats.rx_dropped;
+}
+
+unsigned long __connman_service_stats_get_tx_dropped(struct connman_service *service)
+{
+	return service->stats.tx_dropped;
+}
+
+unsigned long __connman_service_stats_get_time(struct connman_service *service)
+{
+	return service->stats.time;
+}
+
 static struct connman_service *get_default(void)
 {
 	struct connman_service *service;
@@ -500,6 +729,18 @@ static void passphrase_changed(struct connman_service *service)
 						DBUS_TYPE_BOOLEAN, &required);
 }
 
+static void login_changed(struct connman_service *service)
+{
+	dbus_bool_t required = service->login_required;
+
+	if (service->path == NULL)
+		return;
+
+	connman_dbus_property_changed_basic(service->path,
+				CONNMAN_SERVICE_INTERFACE, "LoginRequired",
+						DBUS_TYPE_BOOLEAN, &required);
+}
+
 static void apn_changed(struct connman_service *service)
 {
 	dbus_bool_t required;
@@ -721,6 +962,9 @@ static void append_properties(DBusMessageIter *dict, dbus_bool_t limited,
 		connman_dbus_dict_append_basic(dict, "Name",
 					DBUS_TYPE_STRING, &service->name);
 
+	connman_dbus_dict_append_basic(dict, "LoginRequired",
+				DBUS_TYPE_BOOLEAN, &service->login_required);
+
 	switch (service->type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
@@ -915,6 +1159,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (type != DBUS_TYPE_STRING)
 			return __connman_error_invalid_arguments(msg);
 
+		if (service->immutable == TRUE)
+			return __connman_error_not_supported(msg);
+
 		if (__connman_security_check_privilege(msg,
 					CONNMAN_SECURITY_PRIVILEGE_SECRET) < 0)
 			return __connman_error_permission_denied(msg);
@@ -937,6 +1184,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (type != DBUS_TYPE_STRING)
 			return __connman_error_invalid_arguments(msg);
 
+		if (service->immutable == TRUE)
+			return __connman_error_not_supported(msg);
+
 		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
 			return __connman_error_invalid_service(msg);
 
@@ -958,6 +1208,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (type != DBUS_TYPE_STRING)
 			return __connman_error_invalid_arguments(msg);
 
+		if (service->immutable == TRUE)
+			return __connman_error_not_supported(msg);
+
 		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
 			return __connman_error_invalid_service(msg);
 
@@ -977,6 +1230,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (type != DBUS_TYPE_STRING)
 			return __connman_error_invalid_arguments(msg);
 
+		if (service->immutable == TRUE)
+			return __connman_error_not_supported(msg);
+
 		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
 			return __connman_error_invalid_service(msg);
 
@@ -993,6 +1249,8 @@ static DBusMessage *set_property(DBusConnection *conn,
 	} else if (g_str_equal(name, "Nameservers.Configuration") == TRUE) {
 		DBusMessageIter entry;
 		GString *str;
+		int index;
+		const char *gw;
 
 		if (type != DBUS_TYPE_ARRAY)
 			return __connman_error_invalid_arguments(msg);
@@ -1001,6 +1259,12 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (str == NULL)
 			return __connman_error_invalid_arguments(msg);
 
+		index = connman_network_get_index(service->network);
+		gw = __connman_ipconfig_get_gateway(index);
+
+		if (gw && strlen(gw))
+			__connman_service_nameserver_del_routes(service);
+
 		dbus_message_iter_recurse(&value, &entry);
 
 		while (dbus_message_iter_get_arg_type(&entry) == DBUS_TYPE_STRING) {
@@ -1022,6 +1286,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		g_string_free(str, TRUE);
 
+		if (gw && strlen(gw))
+			__connman_service_nameserver_add_routes(service, gw);
+
 		update_nameservers(service);
 		dns_configuration_changed(service);
 
@@ -1126,6 +1393,9 @@ static DBusMessage *clear_property(DBusConnection *conn,
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
 	} else if (g_str_equal(name, "Passphrase") == TRUE) {
+		if (service->immutable == TRUE)
+			return __connman_error_not_supported(msg);
+
 		g_free(service->passphrase);
 		service->passphrase = NULL;
 
@@ -1433,6 +1703,9 @@ static DBusMessage *remove_service(DBusConnection *conn,
 	if (service->type == CONNMAN_SERVICE_TYPE_ETHERNET)
 		return __connman_error_not_supported(msg);
 
+	if (service->immutable == TRUE)
+		return __connman_error_not_supported(msg);
+
 	if (service->favorite == FALSE &&
 			service->state != CONNMAN_SERVICE_STATE_FAILURE)
 		return __connman_error_not_supported(msg);
@@ -1518,6 +1791,16 @@ static DBusMessage *move_after(DBusConnection *conn,
 	return move_service(conn, msg, user_data, FALSE);
 }
 
+static DBusMessage *reset_stats(DBusConnection *conn,
+					DBusMessage *msg, void *user_data)
+{
+	struct connman_service *service = user_data;
+
+	__connman_service_reset_stats(service);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
 static GDBusMethodTable service_methods[] = {
 	{ "GetProperties", "",   "a{sv}", get_properties     },
 	{ "SetProperty",   "sv", "",      set_property       },
@@ -1528,6 +1811,7 @@ static GDBusMethodTable service_methods[] = {
 	{ "Remove",        "",   "",      remove_service     },
 	{ "MoveBefore",    "o",  "",      move_before        },
 	{ "MoveAfter",     "o",  "",      move_after         },
+	{ "ResetCounters", "",   "",      reset_stats        },
 	{ },
 };
 
@@ -1547,6 +1831,9 @@ static void service_free(gpointer user_data)
 
 	g_hash_table_remove(service_hash, service->identifier);
 
+	__connman_service_stats_stop(service);
+	__connman_storage_save_service(service);
+
 	service->path = NULL;
 
 	if (path != NULL) {
@@ -1588,6 +1875,10 @@ static void service_free(gpointer user_data)
 	g_free(service->private_key_file);
 	g_free(service->private_key_passphrase);
 	g_free(service->phase2);
+
+	if (service->stats.timer != NULL)
+		g_timer_destroy(service->stats.timer);
+
 	g_free(service);
 }
 
@@ -1636,6 +1927,10 @@ static void __connman_service_initialize(struct connman_service *service)
 	service->userconnect = FALSE;
 
 	service->order = 0;
+
+	service->stats.valid = FALSE;
+	service->stats.enabled = FALSE;
+	service->stats.timer = g_timer_new();
 }
 
 /**
@@ -1901,6 +2196,15 @@ int __connman_service_indicate_state(struct connman_service *service,
 	service->state = state;
 	state_changed(service);
 
+	if (state == CONNMAN_SERVICE_STATE_ONLINE) {
+		if (service->login_required == TRUE) {
+			service->login_required = FALSE;
+			login_changed(service);
+		}
+
+		connman_timeserver_sync();
+	}
+
 	if (state == CONNMAN_SERVICE_STATE_IDLE) {
 		connman_bool_t reconnect;
 
@@ -1987,6 +2291,19 @@ int __connman_service_indicate_default(struct connman_service *service)
 	return 0;
 }
 
+int __connman_service_request_login(struct connman_service *service)
+{
+	DBG("service %p", service);
+
+	if (service == NULL)
+		return -EINVAL;
+
+	service->login_required = TRUE;
+	login_changed(service);
+
+	return 0;
+}
+
 static connman_bool_t prepare_network(struct connman_service *service)
 {
 	enum connman_network_type type;
@@ -2484,7 +2801,11 @@ static int service_register(struct connman_service *service)
 
 static void service_up(struct connman_ipconfig *ipconfig)
 {
+	struct connman_service *service = connman_ipconfig_get_data(ipconfig);
+
 	connman_info("%s up", connman_ipconfig_get_ifname(ipconfig));
+
+	service->stats.valid = FALSE;
 }
 
 static void service_down(struct connman_ipconfig *ipconfig)
@@ -2494,12 +2815,21 @@ static void service_down(struct connman_ipconfig *ipconfig)
 
 static void service_lower_up(struct connman_ipconfig *ipconfig)
 {
+	struct connman_service *service = connman_ipconfig_get_data(ipconfig);
+
 	connman_info("%s lower up", connman_ipconfig_get_ifname(ipconfig));
+
+	__connman_service_stats_start(service);
 }
 
 static void service_lower_down(struct connman_ipconfig *ipconfig)
 {
+	struct connman_service *service = connman_ipconfig_get_data(ipconfig);
+
 	connman_info("%s lower down", connman_ipconfig_get_ifname(ipconfig));
+
+	__connman_service_stats_stop(service);
+	__connman_storage_save_service(service);
 }
 
 static void service_ip_bound(struct connman_ipconfig *ipconfig)
@@ -2853,6 +3183,7 @@ void __connman_service_update_from_network(struct connman_network *network)
 	connman_uint8_t strength, value;
 	connman_bool_t roaming;
 	GSequenceIter *iter;
+	const char *name;
 
 	DBG("network %p", network);
 
@@ -2863,6 +3194,15 @@ void __connman_service_update_from_network(struct connman_network *network)
 	if (service->network == NULL)
 		return;
 
+	name = connman_network_get_string(service->network, "Name");
+	if (g_strcmp0(service->name, name) != 0) {
+		g_free(service->name);
+		service->name = g_strdup(name);
+		connman_dbus_property_changed_basic(service->path,
+				CONNMAN_SERVICE_INTERFACE, "Name",
+				DBUS_TYPE_STRING, &service->name);
+	}
+
 	strength = connman_network_get_uint8(service->network, "Strength");
 	if (strength == service->strength)
 		goto roaming;
@@ -2912,6 +3252,54 @@ void __connman_service_remove_from_network(struct connman_network *network)
 	__connman_service_put(service);
 }
 
+void __connman_service_stats_update(struct connman_service *service,
+				unsigned int rx_packets, unsigned int tx_packets,
+				unsigned int rx_bytes, unsigned int tx_bytes,
+				unsigned int rx_errors, unsigned int tx_errors,
+				unsigned int rx_dropped, unsigned int tx_dropped)
+{
+	unsigned int seconds;
+	struct connman_stats *stats = &service->stats;
+
+	DBG("service %p", service);
+
+	if (is_connected(service) == FALSE)
+		return;
+
+	if (stats->valid == TRUE) {
+		stats->rx_packets +=
+			rx_packets - stats->rx_packets_last;
+		stats->tx_packets +=
+			tx_packets - stats->tx_packets_last;
+		stats->rx_bytes +=
+			rx_bytes - stats->rx_bytes_last;
+		stats->tx_bytes +=
+			tx_bytes - stats->tx_bytes_last;
+		stats->rx_errors +=
+			rx_errors - stats->rx_errors_last;
+		stats->tx_errors +=
+			tx_errors - stats->tx_errors_last;
+		stats->rx_dropped +=
+			rx_dropped - stats->rx_dropped_last;
+		stats->tx_dropped +=
+			tx_dropped - stats->tx_dropped_last;
+	} else {
+		stats->valid = TRUE;
+	}
+
+	stats->rx_packets_last = rx_packets;
+	stats->tx_packets_last = tx_packets;
+	stats->rx_bytes_last = rx_bytes;
+	stats->tx_bytes_last = tx_bytes;
+	stats->rx_errors_last = rx_errors;
+	stats->tx_errors_last = tx_errors;
+	stats->rx_dropped_last = rx_dropped;
+	stats->tx_dropped_last = tx_dropped;
+
+	seconds = g_timer_elapsed(stats->timer, NULL);
+	stats->time = stats->time_start + seconds;
+}
+
 static int service_load(struct connman_service *service)
 {
 	const char *ident = service->profile;
@@ -3069,6 +3457,7 @@ static int service_load(struct connman_service *service)
 		service->domains = NULL;
 	}
 
+	__connman_service_stats_load(service, keyfile, service->identifier);
 done:
 	g_key_file_free(keyfile);
 
@@ -3222,6 +3611,8 @@ update:
 		g_key_file_remove_key(keyfile, service->identifier,
 							"Domains", NULL);
 
+	__connman_service_stats_save(service, keyfile, service->identifier);
+
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
 	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
